from lemmymodbot.processors import Processor, Content, LemmyHandle, ContentResult
from lemmymodbot.processors.base import ContentType


class PhashProcessor(Processor):

    def execute(self, content: Content, handle: LemmyHandle) -> ContentResult:
        if content.type != ContentType.POST_LINK:
            return ContentResult.nothing()

        phash = handle.database.url_exists(content.content)
        if phash is not None:
            return self._warn_user(handle, phash)

        phash = handle.fetch_image(content.content)[1]
        if phash is None:
            return ContentResult.nothing()
        
        if handle.database.phash_exists(phash):
            handle.database.add_phash(content.content, phash)
            return self._warn_user(handle, phash)

        handle.database.add_phash(content.content, phash)

        return ContentResult([], {"phash": phash})

    def _warn_user(self, handle: LemmyHandle, phash: str) -> ContentResult:
        posts = handle.database.get_post_links_by_phash(phash)
        other_posts = ', '.join([f"[link]({post})" for post in posts])
        handle.post_comment(f"This post appears to be a duplicate of the following post{ '' if len(posts) == 1 else 's' }: {other_posts}. "
                            f"This could be a false positive (beep boop I am a robot).")
        return ContentResult([], {"phash": phash})



import torch

from lemmymodbot.ml import ToxicityClassifier
from lemmymodbot.processors import Processor, Content, LemmyHandle, ContentResult
from lemmymodbot.processors.base import ContentType


class ToxicityProcessor(Processor):
    classifier: ToxicityClassifier
    uncertainty_allowance = 0.2

    def __init__(self, classifier: ToxicityClassifier = ToxicityClassifier()):
        self.classifier = classifier

    def setup(self) -> None:
        self.classifier.setup()

    def execute(self, content: Content, handle: LemmyHandle) -> ContentResult:
        if content.type == ContentType.POST_LINK:
            return ContentResult.nothing()

        local_flags = []
        preds, preds2, preds3 = self.classifier.classify(content)
        print('\n\n' + content.content)
        print(f'{preds}', (1 - torch.argmax(preds)).item())
        if preds[0].item() < 0.2 and preds[1].item() < 0.2:
            print("Low values^")
        if abs(preds[0].item() - preds[1].item()) < self.uncertainty_allowance:
            print("Close values^")
        if preds2 == 1 and preds3 >= self.uncertainty_allowance:
            print("Toxic^")
            local_flags.append('potentially toxic')

        return ContentResult(local_flags, {
            "toxicity": preds[0].item(),
            "non_toxicity": preds[1].item(),
        })



from typing import Any

import numpy as np
import pandas as pd
import torch
import torchtext
from pandas import DataFrame
from torch.types import Device
from torchtext.vocab import Vocab

from .bag_of_words import build_bow_model, BagOfWords


class ToxicityClassifier:
    train_filename: str
    rebuild_model: bool
    device: Device
    tokenizer: Any
    all_data: DataFrame
    vocab: Vocab
    vocab_size: int
    model: BagOfWords

    def __init__(self, train_filename: str = "training/train.tsv", rebuild_model: bool = True):
        self.train_filename = train_filename
        self.rebuild_model = rebuild_model

    def setup(self) -> None:
        if self.rebuild_model:
            build_bow_model()

        # initialise deep neural network model
        """
        Load the data and set up the tokenizer and dataset for creating/training/using the model.

        :param train_filename: file path and location for the dataset
        :return len(vocab): the length of the vocabulary
        """

        # global tokenizer, vocab, device

        torch.set_default_dtype(torch.float64)

        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # load training data for vocab
        self.all_data = pd.read_csv(self.train_filename, sep='\t', lineterminator='\n')

        self.tokenizer = torchtext.data.utils.get_tokenizer("basic_english")
        self.all_data['tokens'] = self.all_data.apply(lambda row: self.tokenizer(row['comment']), axis=1)

        # build a vocabulary
        self.vocab = torchtext.vocab.build_vocab_from_iterator(self.all_data['tokens'],
                                                               min_freq=5,
                                                               specials=['<unk>', '<pad>'])
        self.vocab.set_default_index(self.vocab['<unk>'])

        self.vocab_size = len(self.vocab)
        self.model = BagOfWords(vocab_size=self.vocab_size)
        self.model.load_state_dict(torch.load("./data/model.mdl"))
        self.model.eval()

    def classify(self, content):
        my_tokens = {}

        my_tokens['tokens'] = self.tokenizer(content.content)
        my_numericals = {'ids': self.numericalize_data(my_tokens)}
        my_multi_hot = self.multi_hot_data(my_numericals)
        inputs = torch.tensor(my_multi_hot).to(self.device)
        # if inputs.dtype != torch.float64:
        #     inputs = inputs.float()
        preds = self.model(inputs)
        preds2 = (1 - torch.argmax(preds, dim=-1)).item()
        preds3 = abs(preds[0].item() - preds[1].item())
        return (preds, preds2, preds3)

    def numericalize_data(self, example):
        """ Convert tokens into vocabulary index."""
        ids = [self.vocab[token] for token in example['tokens']]
        return ids

    def multi_hot_data(self, example):
        """ Convert to a mult-hot list for training/assessment"""
        num_classes = len(self.vocab)
        encoded = np.zeros((num_classes,))
        encoded[example['ids']] = 1
        return encoded



from .base import Processor, LemmyHandle, Content, ContentResult, ContentType
from .user_processor import UserProcessor
from .toxicity_processor import ToxicityProcessor
from .blacklist_processor import BlacklistProcessor
from .phash_processor import PhashProcessor
from .title_conformity_processor import TitleConformityProcessor
from .mime_processor import MimeWhitelistProcessor, MimeBlacklistProcessor
from .account_age_processor import AccountAgeProcessor
from .spam_image_processor import SpamImageProcessor



from io import BytesIO

import requests
from PIL import Image, UnidentifiedImageError
import imagehash


def fetch_image(url: str) -> (Image, str):
    try:
        img = Image.open(BytesIO(requests.get(url).content))
        return img, str(imagehash.phash(img))
    except UnidentifiedImageError:
        return None, None


from typing import List, Callable

from PIL.Image import Image

from lemmymodbot.database import Database
from lemmymodbot.helpers import fetch_image


class SpamImageBootstrapper:
    database: Database
    fetch_image: Callable[[str], tuple[Image, str]]

    def __init__(self, database: Database, fetch_image: Callable[[str], tuple[Image, str]] = fetch_image):
        self.database = database
        self.fetch_image = fetch_image

    def setup(self, images: List[str]):
        for image in images:
            is_url = image.startswith("http")
            if is_url:
                # Check if url has already been added
                phash = self.database.url_exists(image)
                if phash is not None:
                    if self.database.phash_exists(phash, True):
                        continue

            phash = self.fetch_image(image)[1] if is_url else image
            if phash is not None:
                if self.database.phash_exists(phash, True):
                    continue

                self.database.add_phash(image if is_url else "", phash, True)



import re

from lemmymodbot.processors import Processor, Content, LemmyHandle, ContentResult
from lemmymodbot.processors.base import ContentType


class TitleConformityProcessor(Processor):
    message: str

    def __init__(self, regex: str, message: str):
        self.pattern = re.compile(regex)
        self.message = message

    def execute(self, content: Content, handle: LemmyHandle) -> ContentResult:
        if content.type != ContentType.POST_TITLE:
            return ContentResult.nothing()

        if not self.pattern.match(content.content):
            handle.post_comment(self.message)

        return ContentResult.nothing()



from dataclasses import dataclass
from datetime import datetime
from typing import List, Any, Optional, Union, Dict

from dateutil import parser
from plemmy import LemmyHttp
from plemmy.objects import Person
from plemmy.views import PostView, CommentView

from lemmymodbot.helpers import fetch_image
from lemmymodbot.database import Database

import requests
from PIL import Image


@dataclass
class AccountDetails:
    age: float


class LemmyHandle:
    content_footer = "\n\nMod bot (with L plates)"

    def __init__(self, lemmy: LemmyHttp, elem: Union[PostView, CommentView], person: Person, database: Database, config, matrix_facade):
        self.elem = elem
        self.person = person
        self.lemmy = lemmy
        self.lemmy_http = lemmy
        self.database = database
        self.config = config
        self.matrix_facade = matrix_facade

    def send_message_to_author(self, content: str):
        if self.config.debug_mode:
            print(f"{content}")
            return

        actor_id = self.elem.creator.actor_id
        self.lemmy_http.create_private_message(f"{content}{self.content_footer}", actor_id)

    def post_comment(self, content: str):
        if self.config.debug_mode:
            print(f"{content}")
            return

        self.lemmy.create_comment(
            f"{content}{self.content_footer}",
            self.elem.post.id,
            parent_id=self.elem.comment.id if isinstance(self.elem, CommentView) else None
        )

    def remove_thing(self, reason: str):
        if self.config.debug_mode:
            print(f"Remove {reason}")
            return
        if isinstance(self.elem, PostView):
            self.lemmy.remove_post(
                self.elem.post.id,
                True,
                reason
            )
        elif isinstance(self.elem, CommentView):
            self.lemmy.remove_comment(
                self.elem.comment.id,
                True,
                reason
            )

    def _get_url(self) -> Optional[str]:
        if not(isinstance(self.elem, PostView)) or self.elem.post.url is None:
            return None
        return self.elem.post.url

    def fetch_image(self, url: str = None) -> (Image, str):
        if url is None:
            url = self._get_url()
        return fetch_image(url)

    def fetch_content(self, url: str = None) -> (bytes, Dict[str, str]):
        if url is None:
            url = self._get_url()

        cont = requests.get(
            url,
            allow_redirects=True,
            headers={
                "Accepts": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:109.0) Gecko/20100101 Firefox/119.0"
            }
        )
        return cont.content, cont.headers

    def send_message(self, message):
        if self.matrix_facade is None:
            return
        self.matrix_facade.send_message(
            self.config.matrix_config.room_id,
            f"{message}{self.content_footer}"
        )

    def get_account_details(self) -> AccountDetails:
        return AccountDetails(
            (datetime.now() - parser.parse(self.person.published)).total_seconds()
        )


class ContentType:
    POST_TITLE = 0
    POST_BODY = 1
    POST_LINK = 2
    COMMENT = 3


class Content:
    community: str
    content: str
    actor_id: str
    link_to_content: str
    type: ContentType

    def __init__(self, community: str, content: str, actor_id: str, link_to_content: str, type: ContentType):
        self.community = community
        self.content = content
        self.actor_id = actor_id
        self.link_to_content = link_to_content
        self.type = type


class ContentResult:
    flags: List[str]
    extras: Optional[Any]
    was_deleted: bool

    def __init__(self, flags: List[str], extras: Optional[Any], was_deleted: bool = False):
        self.flags = flags
        self.extras = extras
        self.was_deleted = was_deleted

    @staticmethod
    def nothing():
        return ContentResult([], None)


class Processor:

    def setup(self) -> None:
        pass

    def execute(self, content: Content, handle: LemmyHandle) -> ContentResult:
        return ContentResult.nothing()






from typing import List

from lemmymodbot.processors import Processor, Content, LemmyHandle, ContentResult
from lemmymodbot.processors.base import ContentType


class UserProcessor(Processor):
    user_watch_list: List[str]

    def __init__(self, user_watch_list: List[str]):
        self.user_watch_list = user_watch_list

    def execute(self, content: Content, handle: LemmyHandle) -> ContentResult:
        if content.type == ContentType.POST_LINK:
            return ContentResult.nothing()

        if content.actor_id in self.user_watch_list:
            return ContentResult(['user_watch_list'], None)
        return ContentResult.nothing()



from dataclasses import dataclass

import requests
from pykson import JsonObject, StringField, BooleanField, Pykson


@dataclass
class PhotoDNARequest(JsonObject):
    data_representation = StringField(serialized_name="DataRepresentation")
    value = StringField(serialized_name="Value")


@dataclass
class PhotoDNAResponse:
    is_match = BooleanField(serialized_name="IsMatch")


class PhotoDNAClient:
    base_url: str = "https://api.microsoftmoderator.com/photodna"
    api_key: str
    pykson: Pykson

    def __init__(self, api_key: str):
        self.api_key = api_key
        self.pykson = Pykson()

    def _do_match(self, request: PhotoDNARequest) -> PhotoDNAResponse:
        return self.pykson.from_json(requests.post(
            f"{self.base_url}/Match",
            data=self.pykson.to_json(request),
            headers={
                "Ocp-Apim-Subscription-Key": self.api_key
            }
        ).content.decode("utf-8"), PhotoDNAResponse)

    def check_url(self, url: str) -> PhotoDNAResponse:
        request = PhotoDNARequest()
        request.data_representation = "URL"
        request.value = url
        return self._do_match(
            request
        )



from sqlalchemy import select
from sqlalchemy.orm import Session

from lemmymodbot.data.base import CurrentPage, session_scope


class MonitorPersistence:

    def _get_page_entry(self, community_name: str, session: Session) -> CurrentPage:
        return (session.execute(select(CurrentPage).filter(CurrentPage.community_name == community_name))
                .scalar_one_or_none())

    def get_current_page(self, community_name: str) -> int:
        pass

    def set_current_page(self, community_name: str, page_number: int):
        pass


class PostMonitorPersistence(MonitorPersistence):

    def get_current_page(self, community_name: str) -> int:
        with session_scope() as session:
            result = self._get_page_entry(community_name, session)
            if result is None:
                return 1
            return result.post_page

    def set_current_page(self, community_name: str, page_number: int):
        with session_scope() as session:
            current = self._get_page_entry(community_name, session)
            if current is None:
                session.add(CurrentPage(
                    community_name=community_name,
                    post_page=page_number,
                    comment_page=1
                ))
            else:
                current.post_page = page_number


class CommentMonitorPersistence(MonitorPersistence):

    def get_current_page(self, community_name: str) -> int:
        with session_scope() as session:
            result = self._get_page_entry(community_name, session)
            if result is None:
                return 1
            return result.comment_page

    def set_current_page(self, community_name: str, page_number: int):
        with session_scope() as session:
            current = self._get_page_entry(community_name, session)
            if current is None:
                session.add(CurrentPage(
                    community_name=community_name,
                    post_page=1,
                    comment_page=page_number
                ))
            else:
                current.comment_page = page_number



from typing import List, Dict

from matrix_client.client import MatrixClient


class MatrixFacade:
    server: str
    account: str
    password: str

    def __init__(self, server: str, account: str, password: str):
        self.server = server
        self.account = account
        self.password = password

    def send_message(self, room_id: str, content: str):
        client = MatrixClient(self.server)
        client.login(username=self.account, password=self.password, sync=True)
        room = client.join_room(room_id)
        room.send_text(text=content)
        client.logout()

    def report(self, flags: List[str], extras: Dict[str, str], content: str) -> str:
        matrix_message = '\n\nMod bot (with L plates) : ' + ', '.join(flags)
        matrix_message = matrix_message + '\n' + str(extras)
        matrix_message = matrix_message + '\n' + content
        return matrix_message



from typing import List, Optional

from magic import Magic

from lemmymodbot.processors import Processor, Content, LemmyHandle, ContentResult
from lemmymodbot.processors.base import ContentType


class MimeProcessor(Processor):
    match: List[str]
    # Refers to type, as in type/subtype
    type: Optional[List[str]]

    def execute(self, content: Content, handle: LemmyHandle) -> ContentResult:
        if content.type != ContentType.POST_LINK:
            return ContentResult.nothing()

        file, headers = handle.fetch_content(content.content)
        mime = Magic(mime=True).from_buffer(file)
        components = mime.split("/")
        return self.handle(mime in self.match, self.type is not None and components[0] in self.type, mime)

    def handle(self, in_match_list: bool, in_type_list: bool, mime: str) -> ContentResult:
        pass


class MimeWhitelistProcessor(MimeProcessor):

    def __init__(self, whitelist: Optional[List[str]] = None, type_whitelist: Optional[List[str]] = None):
        if whitelist is None:
            whitelist = []
        self.match = whitelist
        self.type = type_whitelist

    def handle(self, in_match_list: bool, in_type_list: bool, mime: str) -> ContentResult:
        return ContentResult([] if in_match_list or in_type_list else ["not_whitelisted_mime"], {"mime": mime})


class MimeBlacklistProcessor(MimeProcessor):
    def __init__(self, blacklist: Optional[List[str]] = None):
        if blacklist is None:
            blacklist = []
        self.match = blacklist

    def handle(self, in_match_list: bool, in_type_list: bool, mime: str) -> ContentResult:
        return ContentResult(["blacklisted_mime"] if in_match_list else [], {"mime": mime})
